module sinm_cste
  implicit none
  integer, parameter :: pr = selected_real_kind( p = 12 )
  integer, parameter :: name_length = 5
  real ( kind = pr ) :: pi
  real ( kind = pr ) :: pi2,  & !  pi*2
       &                t13,  & !  1 / 3
       &                t23,  & !  2 / 3
       &                t43,  & !  4 / 3
       &                t53,  & !  5 / 3
       &                t73,  & !  4 / 3
       &                dtpi2   !  3 * pi**2 / 2

contains

  !
  !!
  !

  subroutine setcste()
    !-------------------------------------------
    !
    ! Purpose:
    !  Sets several usefull constantes...
    ! Input:
    ! Output:
    !
    ! Certified: 23/06/2003
    !
    !-------------------------------------------
    implicit none
    pi = 4 * atan(1.0_pr)  !  pi
    pi2 = pi * pi          !  pi**2
    t13 = 1.0_pr / 3.0_pr  !  1 / 3
    t23 = 2 * t13          !  2 / 3
    t43 = 4 * t13          !  4 / 3
    t53 = 5 * t13          !  5 / 3
    t73 = 7 * t13          !  7 / 3
    dtpi2 = 1.5_pr * pi2   !  3 / 2 * pi**2
    !
  end subroutine setcste

  !
  !!
  !

end module sinm_cste
module sinm_skyrme
  use sinm_cste
  implicit none

  !
  !!
  !

  real ( kind = pr ) :: t0, t1, t2, t3a, t3b, t3c
  real ( kind = pr ) :: x0, x1, x2, x3a, x3b, x3c
  real ( kind = pr ) :: gama, gamb, gamc, w0, w1, tt, tu
  real ( kind = pr ) :: hb2, thetas, thetav
  logical :: j2terms, xtls

  !
  !!
  !

contains

  !
  !!
  !

  subroutine set_theta()
    implicit none
    !
    if ( thetas == 0.0_pr ) then
       thetas = 3 * t1 + ( 5 + 4 * x2 ) * t2
    end if
    if ( thetav == 0.0_pr ) then
       thetav = t1 * ( 2 + x1 ) + t2 * ( 2 + x2 )
    end if
    !
  end subroutine set_theta

  !
  !!
  !

  subroutine read_force(force)
    implicit none
    character ( len = * ), intent(in) :: force
    character ( len = name_length ) :: name
    real ( kind = pr ) :: sigma_n, sigma_d, sigma2_n, sigma2_d, w
    real ( kind = pr ) :: t3, x3, hb2m, t4, x4, sigma2, sigma, b4, b4p
    logical :: cm1
    namelist /skf/ name, t0, t1, t2, t3, t4, x0, x1, x2, x3, x4, w, tt, tu,  &
         sigma, sigma2, j2terms, sigma_n, sigma_d, hb2m, cm1, &
         sigma2_n, sigma2_d, b4, b4p
    !
    open( unit = 14, file = "forces.param", status = "old" )
    do
       call set_def()
       read( 14, nml = skf )
       if ( name == "FIN" .or. name == "fin" .or. name == "Fin" ) then
          print '(" Unknown force...")'
          stop
       end if
       if ( name == force ) exit
    end do
    close(14)
    if ( sigma_n  * sigma_d /= 0.0_pr  .and. sigma  == 0.0_pr )  &
         sigma  = sigma_n  / sigma_d
    if ( sigma2_n * sigma2_d /= 0.0_pr .and. sigma2 == 0.0_pr )  &
         sigma2 = sigma2_n / sigma2_d
    if ( abs(b4) + abs(b4p) /= 0.0_pr ) stop "Not fully implemented yet..."
    write( *, nml = skf )
    t3a = t3
    x3a = x3
    gama = sigma
    t3b = t4
    x3b = x4
    gamb = sigma2
    t3c = 0.0_pr
    x3c = 0.0_pr
    gamc = 0.0_pr
    w0 = w
    w1 = 0.0_pr
    hb2 = hb2m
    xtls = .false.
    !
  contains

    !
    !!
    !

    subroutine set_def()
      implicit none
      !
      t0 = -2488.913_pr * 0
      t1 =   486.818_pr * 0
      t2 =  -546.395_pr * 0
      t3 = 13777.000_pr * 0
      t4 =     0.000_pr * 0
      x0 =     0.834_pr * 0
      x1 =    -0.344_pr * 0
      x2 =    -1.000_pr * 0
      x3 =     1.354_pr * 0
      x4 =     0.000_pr * 0
      w   = 123.000_pr * 0
      b4  =   0.000_pr * 0
      b4p =   0.000_pr * 0
      tt  =   0.000_pr * 0
      tu  =   0.000_pr * 0
      sigma = 0.000_pr
      sigma2 = 0.000_pr
      sigma_n  = 1.0_pr * 0
      sigma_d  = 6.0_pr * 0
      sigma2_n = 0.0_pr * 0
      sigma2_d = 6.0_pr * 0
      j2terms  = .false.
      cm1      = .true.
      hb2m     = 20.73553_pr
      !
    end subroutine set_def

    !
    !!
    !

  end subroutine read_force

  !
  !!
  !

end module sinm_skyrme
module sinm_system
  use sinm_cste
  implicit none
  !
  real ( kind = pr ) :: yp, rho0(3), kf(3), m_star(2), kappa_v
  !
end module sinm_system
module sinm_geometry
  use sinm_cste
  implicit none
  !
  integer :: npt, nkt, nkz, n3, m_kz
  !
  real ( kind = pr ) :: r_box, h, h_12, hh_12, hh_24, hh12, h_60
  !
  !! coef. for derivation and integration
  !
  real ( kind = pr ) :: c_d1(3), c_d2(3), h23, h43, h_3
  !
  real ( kind = pr ) :: xf
  character ( len = 1 ) :: integration = "s"
  !
  real ( kind = pr ), dimension(:), allocatable :: z
  !
  !! Densities
  !
  real ( kind = pr ), dimension(:,:), allocatable :: rho,   tau,   spin
  real ( kind = pr ), dimension(:,:), allocatable :: drho,  dtau,  dspin
  real ( kind = pr ), dimension(:,:), allocatable :: d2rho, d2tau, d2spin
  !
  !! Energy densities
  !
  real ( kind = pr ), dimension(:,:), allocatable :: energy
  !
  !! Effective mass
  !
  real ( kind = pr ), dimension(:,:), allocatable :: hb2m, dhb2m, d2hb2m
  !
  !! Fields
  !
  real ( kind = pr ), dimension(:,:), allocatable :: u_cent, u_so


contains

  !
  !!
  !

  subroutine set_geometry()
    implicit none
    !
    n3 = npt + 3
    r_box = h * npt
    h_12 = 12 * h
    hh_12 = 12 * h * h
    hh_24 = 24 * h * h
    hh12 = h * h / 12
    h_60 = h * 60
    c_d1 = (/  45.0_pr, -9.0_pr,  1.0_pr /) / h_60
    c_d2 = (/ -30.0_pr, 16.0_pr, -1.0_pr /) / hh_12
    h23 = ( 2 * h ) / 3
    h43 = ( 4 * h ) / 3
    h_3 = h / 3
    !
  end subroutine set_geometry

  !
  !!
  !

  subroutine alloc_mem()
    implicit none
    integer :: i, ifail
    !
    allocate( rho(3,-n3:n3),   tau(3,-n3:n3),   spin(3,-n3:n3),         &
         &    drho(3,-n3:n3),  dtau(3,-n3:n3),  dspin(3,-n3:n3),        &
         &    d2rho(3,-n3:n3), d2tau(3,-n3:n3), d2spin(3,-n3:n3),       &
         &    hb2m(2,-n3:n3),  dhb2m(2,-n3:n3), d2hb2m(2,-n3:n3),       &
         &    z(-n3:n3),                                                &
         &    u_cent(2,-npt:npt), u_so(2,-npt:npt),                     &
         &    energy(9,-n3:n3),                                         &
         &    stat = ifail )
    if ( ifail /= 0 ) stop 'Memory...'
    !
    z = (/ ( i * h, i = - n3, n3 ) /)
    !
    !! Flush allocatted arraies
    !
    rho    = 0.0_pr
    tau    = 0.0_pr
    spin   = 0.0_pr
    drho   = 0.0_pr
    dtau   = 0.0_pr
    dspin  = 0.0_pr
    d2rho  = 0.0_pr
    d2tau  = 0.0_pr
    d2spin = 0.0_pr
    hb2m   = 0.0_pr
    dhb2m  = 0.0_pr
    d2hb2m = 0.0_pr
    u_cent = 0.0_pr
    u_so   = 0.0_pr
    energy = 0.0_pr
    !
  end subroutine alloc_mem

  !
  !!
  !

  subroutine free_mem()
    implicit none
    deallocate( rho, tau, spin, drho, dtau, dspin, &
         d2rho, d2tau, d2spin, hb2m, dhb2m, d2hb2m, z, u_cent, u_so, &
         energy )
  end subroutine free_mem

  !
  !!
  !

end module sinm_geometry
module sinm_io
  use sinm_cste
  use sinm_geometry
  implicit none

  integer, parameter :: iout = 10
  logical :: gfx

  !interface writeoutwf
  !   module procedure writeoutwf
  !end interface

  interface printoutdensities
     module procedure printoutdensities
  end interface

  interface printoutfields
     module procedure printoutfields
  end interface

  interface print_param
     module procedure print_param
  end interface

  interface print_inm
     module procedure print_inm
  end interface

contains

  !
  !!
  !

  !subroutine writeoutwf( it, iz, s, f, df )
  !  implicit none
  !  integer, intent(in) :: it, iz, s
  !  real ( kind = pr ), intent(in), dimension(-n:n) :: f, df
  !  character ( len = 2 ) :: strt, strz
  !  character ( len = 1 ) :: strs
  ! integer :: i
  ! !
  ! if ( it < 10 ) then
  !    write( strt, '("0",i1)' ) it
  ! else
  !    write( strt, '(i2)' ) it
  ! end if
  !
  ! if ( iz < 10 ) then
  !    write( strz, '("0",i1)' ) iz
  ! else
  !    write( strz, '(i2)' ) iz
  ! end if
  !
  ! strs = "u"
  ! if ( s < 0 ) strs = "d"
  ! open( unit = 10, file = "out/wf_"//strz//"_"//strt//"_"//strs//".gfx", &
  !      status = "unknown" )
  ! do i = -n, n
  !    write(10,'(f8.3,2e18.7)' ) z(i), f(i), df(i)
  ! end do
  ! close(10)
  !end subroutine writeoutwf

  !
  !!
  !

  subroutine printoutdensities(w)
    use sinm_system
    use sinm_skyrme, only : thetas, thetav, hb2
    implicit none
    integer, intent(in), optional :: w
    character ( len = 4 ) :: sw
    integer :: i
    !
    sw = ""
    if ( present(w) ) write( sw, '(i4)' ) w
    !
    open( unit = 20, file = 'out/rho'//trim(adjustl(sw))//'.gfx', &
         status = 'unknown' )
    open( unit = 21, file = 'out/tau'//trim(adjustl(sw))//'.gfx', &
         status = 'unknown' )
    open( unit = 22, file = 'out/spin'//trim(adjustl(sw))//'.gfx', &
         status = 'unknown' )
    write( 20, '("#  z",10x,"rho_n",11x,"rho_p",11x,"rho_0",11x,"rho_inm")' )
    write( 21, '("#  z",10x,"tau_n",11x,"tau_p",11x,"tau_0",10x,&
         &"tau_n_TF",8x,"tau_p_TF",8x,"tau_n_inm",7x,"tau_p_inm")' )
    write( 22, '("#  z",10x,"J_n",13x,"J_p",13x,"J_0")' )
    do i = -npt, npt
       write(20,'(f7.3,4e16.6e3)') z(i), rho(1:3,i), 2*kf(3)**3/3/pi**2
       write(21,'(f7.3,7e16.6e3)') z(i), tau(1:3,i),        &
            ( 3 * pi**2 * rho(1:2,i) )**t53 / 5 / pi**2,  &
            kf(1:2)**5 / 5 / pi**2
       write(22,'(f7.3,5e16.6e3)') z(i), spin(1:3,i)
    end do
    close(20)
    close(21)
    close(22)
    !
    !! m*
    !
    open( unit = 20, file = 'out/masses'//trim(adjustl(sw))//'.gfx', &
         status = 'unknown' )
    write( 20, '("#  z"9x,"m_n*",8x,"m_p*",6x,"m_n*_inm",4x,"m_p*_inm")' )
    do i = -npt, npt
       write(20,'(f7.3,4f12.6)') z(i), 1 / ( 1  &
            + rho(3,i) * thetav / 8 / hb2       &
            + rho(1,i) * ( thetas - 2 * thetav ) / 8 / hb2 ),   &
            1 / ( 1 + rho(3,i) * thetav / 8 / hb2               &
            + rho(2,i) * ( thetas - 2 * thetav ) / 8 / hb2 ),   &
            m_star
    end do
    close(20)
    !
    !! Derivatives
    !
    open( unit = 20, file = 'out/drho'//trim(adjustl(sw))//'.gfx', &
         status = 'unknown' )
    open( unit = 21, file = 'out/dtau'//trim(adjustl(sw))//'.gfx', &
         status = 'unknown' )
    open( unit = 22, file = 'out/dspin'//trim(adjustl(sw))//'.gfx', &
         status = 'unknown' )
    write( 20, '("#  z",9x,"drho_n",10x,"drho_p",10x,"drho_0")' )
    write( 21, '("#  z",9x,"dtau_n",10x,"dtau_p",10x,"dtau_0")' )
    write( 22, '("#  z",10x,"dJ_n",12x,"dJ_p",12x,"dJ_0")' )
    do i = -npt, npt
       write(20,'(f7.3,4e16.6)') z(i), drho(1:3,i)
       write(21,'(f7.3,5e16.6)') z(i), dtau(1:3,i)
       write(22,'(f7.3,5e16.6)') z(i), dspin(1:3,i)
    end do
    close(20)
    close(21)
    close(22)
    !
    open( unit = 20, file = 'out/d2rho'//trim(adjustl(sw))//'.gfx', &
         status = 'unknown' )
    open( unit = 21, file = 'out/d2tau'//trim(adjustl(sw))//'.gfx', &
         status = 'unknown' )
    open( unit = 22, file = 'out/d2spin'//trim(adjustl(sw))//'.gfx', &
         status = 'unknown' )
    write( 20, '("#  z",9x,"d2rho_n",9x,"d2rho_p",9x,"d2rho_0")' )
    write( 21, '("#  z",9x,"d2tau_n",9x,"d2tau_p",9x,"d2tau_0")' )
    write( 22, '("#  z",10x,"d2J_n",11x,"d2J_p",11x,"d2J_0")' )
    do i = -npt, npt
       write(20,'(f7.3,4e16.6)') z(i), d2rho(1:3,i)
       write(21,'(f7.3,5e16.6)') z(i), d2tau(1:3,i)
       write(22,'(f7.3,5e16.6)') z(i), d2spin(1:3,i)
    end do
    close(20)
    close(21)
    close(22)
    !
  end subroutine printoutdensities

  !
  !!
  !

  subroutine printoutfields(w)
    use sinm_skyrme
    implicit none
    integer, intent(in), optional :: w
    character ( len = 4 ) :: sw
    integer :: i, iso
    !
    sw = ""
    if ( present(w) ) write( sw, '(i4)' ) w
    !
    do iso = 1, 2
       if ( iso == 1 ) then
          open( unit = iout, &
               file = 'out/fields_n'//trim(adjustl(sw))//'.gfx', &
               status = 'unknown' )
       else
          open( unit = iout, &
               file = 'out/fields_p'//trim(adjustl(sw))//'.gfx', &
               status = 'unknown' )
       end if
       do i = -npt, npt
          write( iout, '(1x,f6.2,3e15.6)' ) z(i), hb2 / hb2m(iso,i), &
               u_cent(iso,i), u_so(iso,i)
       end do
       close(iout)
    end do
  end subroutine printoutfields

  !
  !!
  !

  subroutine line()
    implicit none
    print '(72("="))'
  end subroutine line

  !
  !!
  !


  subroutine print_param()
    use sinm_skyrme
    implicit none
    !
    print '(" * Skyrme force parameters:")'
    print '(4x,"t0 =",f10.3,4x,"t1 =",f10.3,4x,"t2 =",f10.3)',  &
         t0, t1, t2
    print '(4x,"x0 =",f10.3,4x,"x1 =",f10.3,4x,"x2 =",f10.3)',  &
         x0, x1, x2
    print '(4x,"DD:  ",&
         &   "t3a  =",f10.3,3x,"x3a  =",f10.3,3x,"gama = ",f10.7,/,9x, &
         &   "t3b  =",f10.3,3x,"x3b  =",f10.3,3x,"gamb = ",f10.7,/,9x, &
         &   "t3c  =",f10.3,3x,"x3c  =",f10.3,3x,"gamc = ",f10.7)',    &
         & t3a, x3a, gama, &
         & t3b, x3b, gamb, &
         & t3c, x3c, gamc
    if ( xtls ) then
       print '(4x,"W0 =",f10.3,3x,"W1 = ",f10.3)', w0, w1
    else
       print '(4x,"W0 =",f10.3)', w0
    end if
    print '(4x,"Theta_s = ",f9.3,3x,"Theta_v = ",f9.3)',  &
         & thetas, thetav
    print '(4x,"hbar2/2m =",f10.6)', hb2
    !
  end subroutine print_param

  !
  !!
  !

  subroutine print_inm( av, ai, kinf, r0 )
    use sinm_system
    implicit none
    real ( kind = pr ), intent(in) :: av, ai, kinf, r0
    !
    print '(" * Infinite nuclear matter:")'
    print '(4x,"Yp = ",f6.3,5x,"I**2 = ",f9.6)', yp, ( 1 - 2 * yp )**2
    print '(4x,"rho0  = ",f8.5," + ",f8.5," = ",f7.4," fm-3")', rho0
    print '(4x,"kf(1) = ",f6.3," fm,",4x,"kf(2) = ",f6.3," fm,",&
         &  4x,"kf = ",f6.3," fm")', kf
    print '(4x,"E/A   = ",f7.3," MeV",25x,"ai = ",f6.2," MeV")', av, ai
    print '(4x,"K_inf = ",f7.2," MeV",25x,"r0 = ",f6.3," fm")', kinf, r0
    print '(4x,"m*_n  =",f7.3,8x,"m*_p  =",f7.3,7x,"k_v =",f7.3)',  &
         m_star, kappa_v
    !
  end subroutine print_inm

  !
  !!
  !

  subroutine print_geo()
    implicit none
    !
    print '(" * Numerical parameters:")'
    print '(4x,"Rbox = 2 *",f7.3,5x,"h = ",f7.4,5x,"xfield = ",f6.3)', &
         r_box, h, xf
    print '(4x,"Nkt =",i3,",",4x,"Nkz = 32 *",i2," = ",i3)', &
         nkt, m_kz, nkz
    !
  end subroutine print_geo

end module sinm_io
module sinm_inm
  use sinm_cste
  use sinm_skyrme
  use sinm_system
  implicit none
  real ( kind = pr ), private, save :: f2, f53, f83, c0, c12, &
       &                               c3a, c3b, c3c

  private
  public :: get_inm

  interface get_inm
     module procedure get_inm
  end interface

contains

  !
  !!
  !

  subroutine get_inm( yp, av, ai, kinf, d3e, r0 )
    use sinm_io
    implicit none
    real ( kind = pr ), intent(in) :: yp
    real ( kind = pr ), intent(out) :: av, ai, kinf, d3e, r0
    integer :: i
    real ( kind = pr ) :: r1, r2, r
    real ( kind = pr ) ::  p1, p2, p, rh
    !
    call initinm(yp)
    !
    rh = 0.02_pr
    av = 0.0_pr
    rho0(3) = 0.0_pr
    i = 1
    do
       r1 = i * rh
       !print *, i, r1, energy_inm( r1)
       if ( energy_inm(r1) < 0.0_pr ) exit
       i = i + 1
       if ( r1 > 1.6_pr ) stop 'No bound state in I.N.M. !'
    end do
    !
    do
       r2 = i * rh
       if ( energy_inm(r2) > 0.0_pr ) exit
       i = i + 1
    end do
    p1 = pressure(r1)
    p2 = pressure(r2)
    !print *, p1, p2
    !stop
    do
       r = ( r1 + r2 ) / 2
       p = pressure(r)
       if ( p * p1 < 0.0_pr ) then
          r2 = r
          p2 = p
       else
          r1 = r
          p1 = p
       end if
       if ( abs(p) < 1.e-11_pr ) exit
    end do
    rho0(3) = r
    rho0(1) = ( 1 - yp ) * r
    rho0(2) = yp * r
    !
    kf(1) = ( 3 * pi**2 * rho0(1) )**t13
    kf(2) = ( 3 * pi**2 * rho0(2) )**t13
    kf(3) = ( 1.5_pr * pi**2 * rho0(3) )**t13
    !
    av = energy_inm(rho0(3))
    r0 = ( 0.75_pr / pi / r )**t13
    d3e = 8.0_pr / 45 * hb2 * dtpi2**t23 / r**t73                       &
         - thetas / 72 * dtpi2**t23 / r**t43                            &
         + t3a * ( gama - 1 ) * gama * ( gama + 1 ) / 16 * r**(gama-2)  &
         + t3b * ( gama - 1 ) * gamb * ( gamb + 1 ) / 16 * r**(gamb-2)  &
         + t3c * ( gama - 1 ) * gamc * ( gamc + 1 ) / 16 * r**(gamc-2)
    kinf = compression_modulus(rho0(3))
    if ( yp == 0.5_pr ) then
       ai   = hb2 * dtpi2**t23 / 3 * r**t23                           &
            - t0 / 8 * ( 2 * x0 + 1 ) * r                             &
            - dtpi2**t23 / 24 * ( 3 * thetav - 2 * thetas ) * r**t53  &
            - t3a / 48 * ( 2 * x3a + 1 ) * r**(gama+1)                &
            - t3b / 48 * ( 2 * x3b + 1 ) * r**(gamb+1)                &
            - t3c / 48 * ( 2 * x3c + 1 ) * r**(gamc+1)
    else
       ai   = 0.0_pr
    end if
    !
    !! Effective masses
    !
    m_star(1) = 1  &
         + 1.0_pr / hb2 / 8 * rho0(3)               &
         * ( t1 * ( 2 + x1 ) + t2 * ( 2 + x2 ) )    &
         + 1.0_pr / hb2 / 8 * rho0(1)               &
         * ( t2 * ( 1 + 2 * x2 ) - t1 * ( 1 + 2 * x1 ) )
    m_star(2) = 1  &
         + 1.0_pr / hb2 / 8 * rho0(3)               &
         * ( t1 * ( 2 + x1 ) + t2 * ( 2 + x2 ) )    &
         + 1.0_pr / hb2 / 8 * rho0(2)               &
         * ( t2 * ( 1 + 2 * x2 ) - t1 * ( 1 + 2 * x1 ) )
    m_star = 1 / m_star
    kappa_v = rho0(3) / 8 / hb2 * ( t1 * ( 2 + x1 ) + t2 * ( 2 + x2 ) )
    !
  end subroutine get_inm

  !
  !!
  !

  subroutine initinm(yp)
    implicit none
    real ( kind = pr ), intent(in) :: yp
    !
    f2  = f( 2.0_pr, yp )
    f53 = f( 5.0_pr / 3.0_pr, yp )
    f83 = f( 8.0_pr / 3.0_pr, yp )
    !
    c0  = 2 * ( x0  + 2 ) - ( 2 * x0  + 1 ) * f2
    c3a = 2 * ( x3a + 2 ) - ( 2 * x3a + 1 ) * f2
    c3b = 2 * ( x3b + 2 ) - ( 2 * x3b + 1 ) * f2
    c3c = 2 * ( x3c + 2 ) - ( 2 * x3c + 1 ) * f2
    c12 = ( t1 * ( x1 + 2 ) + t2 * ( x2 + 2 ) ) * f53  &
         &  + ( t2 * ( 2 * x2 + 1 )                    &
         &    - t1 * ( 2 * x1 + 1 ) ) * f83 / 2
    !
  end subroutine initinm

  !
  !!
  !

  function energy_inm(rho) result(res)
    implicit none
    real ( kind = pr ) :: rho
    real ( kind = pr ) :: res
    !
    res = 0.6_pr * hb2 * dtpi2**t23 * rho**t23 * f53     &
         & + t0 / 8 * rho * c0                           &
         & + 3.0_pr / 40 * dtpi2**t23 * rho**t53 * c12   &
         & + t3a / 48 * rho**(gama+1) * c3a              &
         & + t3b / 48 * rho**(gamb+1) * c3b              &
         & + t3c / 48 * rho**(gamc+1) * c3c
    !
  end function energy_inm

  !
  !!
  !

  function compression_modulus(rho) result(res)
    implicit none
    real ( kind = pr ) :: rho
    real ( kind = pr ) :: res
    real ( kind = pr ) :: ga1, gb1, gc1
    !
    ga1 = gama + 1
    gb1 = gamb + 1
    gc1 = gamc + 1
    res = - 1.2_pr * hb2 * dtpi2**t23 * rho**t23 * f53      &
         & + 0.75_pr * dtpi2**t23 * rho**t53 * c12          &
         & + 0.1875_pr * t3a * rho**ga1 * c3a * ga1 * gama  &
         & + 0.1875_pr * t3b * rho**gb1 * c3b * gb1 * gamb  &
         & + 0.1875_pr * t3c * rho**gc1 * c3c * gc1 * gamc
    !
  end function compression_modulus

  !
  !!
  !

  function f( m, yp ) result(res)
    implicit none
    real ( kind = pr ) :: m, yp, res
    res = 2.0_pr**(m-1) * ( yp**m + ( 1 - yp )**m )
  end function f

  !
  !!
  !

  function pressure(rho) result(res)
    implicit none
    real ( kind = pr ) :: rho
    real ( kind = pr ) :: res
    !
    res = rho * ( 0.4_pr * hb2 * dtpi2**t23 * rho**t23 * f53  &
         & + t0 / 8 * rho * c0                                &
         & + 0.125_pr * dtpi2**t23 * rho**t53 * c12           &
         & + t3a * (gama+1) / 48 * rho**(gama+1) * c3a        &
         & + t3b * (gamb+1) / 48 * rho**(gamb+1) * c3b        &
         & + t3c * (gamc+1) / 48 * rho**(gamc+1) * c3c        )
    !
  end function pressure

  !
  !!
  !

end module sinm_inm
module sinm_difint

  use sinm_cste
  implicit none

  interface deriv
     module procedure derivs
     module procedure derivv
  end interface

  interface simpson
     module procedure scasimpson
     module procedure vecsimpson
  end interface

  interface romberg
     module procedure rombergs
     module procedure rombergv
  end interface

contains

  !
  !!
  !

  subroutine derivs( f, df, d2f )
    !-------------------------------------------
    !
    ! Purpose:
    !  Numerical derivation using a 7 points formula (and a
    !  5 points formula for the second derivative).
    ! Input:
    !  f = 1 dimension array of real( kind = pr )
    !  h = step
    ! Output:
    !  df = derivative of f
    !  d2f = second derivative of f (optional)
    !
    ! Certified: 23/06/2003
    !
    !-------------------------------------------
    use sinm_geometry, only : h_12, hh_24, c_d1, c_d2
    implicit none
    real ( kind = pr ), intent(in) :: f(:)
    real ( kind = pr ), intent(inout) :: df(:)
    real ( kind = pr ), intent(inout), optional :: d2f(:)
    integer :: i, nn
    !
    nn = size(f)
    do i = 4, nn - 3
       !df(i) = ( 45 * ( f(i+1) - f(i-1) )   &
       !     &   - 9 * ( f(i+2) - f(i-2) )   &
       !     &         + f(i+3) - f(i-3) ) / h_60
       df(i) =  c_d1(1) * ( f(i+1) - f(i-1) )   &
            & + c_d1(2) * ( f(i+2) - f(i-2) )   &
            & + c_d1(3) * ( f(i+3) - f(i-3) )
       if ( present(d2f) ) &
            d2f(i) = c_d2(1) * f(i) + c_d2(2) * ( f(i+1) + f(i-1) ) &
            + c_d2(3) * ( f(i+2) + f(i-2) )
       !d2f(i) = ( - 30 * f(i) + 16 * ( f(i+1) + f(i-1) ) &
       !- f(i+2) - f(i-2) ) / hh_12
    end do
    !
    df(1) = ( - 25 * f(1) + 48 * f(2) - 36 * f(3) &
         &      + 16 * f(4) - 3 * f(5) ) / h_12
    df(2) = ( - 3 * f(1) - 10 * f(2) + 18 * f(3) &
         &      - 6 * f(4) + f(5) ) / h_12
    df(3) = ( f(1) + 8 * ( f(4) - f(2) ) - f(5) ) / h_12
    !
    df(nn) = ( 25 * f(nn) - 48 * f(nn-1) + 36 * f(nn-2)  &
         &    - 16 * f(nn-3) + 3 * f(nn-4) ) / h_12
    df(nn-1) = ( 3 * f(nn) + 10 * f(nn-1) - 18 * f(nn-2)  &
         &      + 6 * f(nn-3) - f(nn-4) ) / h_12
    df(nn-2) = ( - f(nn) + 8 * ( f(nn-1) - f(nn-3) ) + f(nn-4) ) / h_12
    if ( present(d2f) ) then
       d2f(1) = ( 70 * f(1) - 208 * f(2) + 228 * f(3) &
            &      - 112 * f(4) + 22 * f(5) ) / hh_24
       d2f(2) = ( 22 * f(1) - 40 * f(2) + 12 * f(3) &
            &      + 8 * f(4) - 2 * f(5) ) / hh_24
       d2f(3) = ( - 2 * f(1) + 32 * f(2) - 60 * f(3) &
            &      + 32 * f(4) - 2 * f(5) ) / hh_24
       !
       d2f(nn) = ( 70 * f(nn) - 208 * f(nn-1) + 228 * f(nn-2) &
            &      - 112 * f(nn-3) + 22 * f(nn-4) ) / hh_24
       d2f(nn-1) = ( 22 * f(nn) - 40 * f(nn-1) + 12 * f(nn-2) &
            &      + 8 * f(nn-3) - 2 * f(nn-4) ) / hh_24
       d2f(nn-2) = ( - 2 * f(nn) + 32 * f(nn-1) - 60 * f(nn-2) &
            &      + 32 * f(nn-3) - 2 * f(nn-4) ) / hh_24
    end if
  end subroutine derivs

  !
  !!
  !

  subroutine derivv( f, df, d2f )
    !-------------------------------------------
    !
    ! Purpose:
    !  Numerical derivation using a 7 points formula (and a
    !  5 points formula for the second derivative).
    ! Input:
    !  f = n dimensions array of real( kind = pr )
    !  h = step
    ! Output:
    !  df = derivative of f (same dimension as f)
    !  d2f = second derivative of f (optional)
    !
    ! Certified: 23/06/2003
    !
    !-------------------------------------------
    use sinm_geometry, only : h_12, hh_24, c_d1, c_d2
    implicit none
    real ( kind = pr ), intent(in) :: f(:,:)
    real ( kind = pr ), intent(inout) :: df(:,:)
    real ( kind = pr ), intent(inout), optional :: d2f(:,:)
    integer :: i, nn
    !
    nn = size(f,2)
    do i = 4, nn - 3
       df(:,i) = c_d1(1) * ( f(:,i+1) - f(:,i-1) )   &
            &  + c_d1(2) * ( f(:,i+2) - f(:,i-2) )   &
            &  + c_d1(3) * ( f(:,i+3) - f(:,i-3) )
       if ( present(d2f) ) &
            d2f(:,i) = c_d2(1) * f(:,i)         &
            + c_d2(2) * ( f(:,i+1) + f(:,i-1) ) &
            + c_d2(3) * ( f(:,i+2) + f(:,i-2) )
    end do
    !
    df(:,1) = ( - 25 * f(:,1) + 48 * f(:,2) - 36 * f(:,3) &
         &      + 16 * f(:,4) - 3 * f(:,5) ) / h_12
    df(:,2) = ( - 3 * f(:,1) - 10 * f(:,2) + 18 * f(:,3) &
         &      - 6 * f(:,4) + f(:,5) ) / h_12
    df(:,3) = ( f(:,1) - 8 * f(:,2) + 8 * f(:,4) - f(:,5) ) / h_12
    !
    df(:,nn) = ( 25 * f(:,nn) - 48 * f(:,nn-1) + 36 * f(:,nn-2)  &
         &    - 16 * f(:,nn-3) + 3 * f(:,nn-4) ) / h_12
    df(:,nn-1) = ( 3 * f(:,nn) + 10 * f(:,nn-1) - 18 * f(:,nn-2)  &
         &      + 6 * f(:,nn-3) - f(:,nn-4) ) / h_12
    df(:,nn-2) = ( - f(:,nn) + 8 * f(:,nn-1)  &
         &      - 8 * f(:,nn-3) + f(:,nn-4) ) / h_12
    !
    if ( present(d2f) ) then
       d2f(:,1) = ( 70 * f(:,1) - 208 * f(:,2) + 228 * f(:,3) &
            &      - 112 * f(:,4) + 22 * f(:,5) ) / hh_24
       d2f(:,2) = ( 22 * f(:,1) - 40 * f(:,2) + 12 * f(:,3) &
            &      + 8 * f(:,4) - 2 * f(:,5) ) / hh_24
       d2f(:,3) = ( - 2 * f(:,1) + 32 * f(:,2) - 60 * f(:,3) &
            &      + 32 * f(:,4) - 2 * f(:,5) ) / hh_24
       !
       d2f(:,nn) = ( 70 * f(:,nn) - 208 * f(:,nn-1) + 228 * f(:,nn-2) &
            &      - 112 * f(:,nn-3) + 22 * f(:,nn-4) ) / hh_24
       d2f(:,nn-1) = ( 22 * f(:,nn) - 40 * f(:,nn-1) + 12 * f(:,nn-2) &
            &      + 8 * f(:,nn-3) - 2 * f(:,nn-4) ) / hh_24
       d2f(:,nn-2) = ( - 2 * f(:,nn) + 32 * f(:,nn-1) - 60 * f(:,nn-2) &
            &      + 32 * f(:,nn-3) - 2 * f(:,nn-4) ) / hh_24
    end if
  end subroutine derivv

  !
  !!
  !

  function scasimpson(f) result(res)
    !-------------------------------------------
    !
    ! Purpose:
    !  Numerical using Simpson method for a scalar real function.
    ! Input:
    !  f = 1 dimension array of real( kind = pr )
    !  h = step
    ! Output:
    !  res = integral of f
    !
    ! Certified: 23/06/2003
    !
    !-------------------------------------------
    use sinm_geometry, only : h23, h43, h_3
    implicit none
    real ( kind = pr ), intent(in) :: f(:)
    real ( kind = pr ) :: res
    integer :: nn, i
    !
    nn = size(f)
    if ( nn == 2 * ( nn / 2 ) ) stop 'Simpson...'
    res = h_3 * ( f(1) - f(nn) )
    do i = 2, nn - 1, 2
       res = res + h43 * f(i) + h23 * f(i+1)
       !res = res + 4 * f(i) + 2 * f(i+1)
    end do
    !res = h / 3 * res
    !
  end function scasimpson

  !
  !!
  !

  function vecsimpson( f, h ) result(res)
    !-------------------------------------------
    !
    ! Purpose:
    !  Numerical using Simpson method for a vector real function.
    ! Input:
    !  f = n dimension array of real( kind = pr )
    !  h = step
    ! Output:
    !  res = n dimensions array, integral of f
    !
    ! Certified: 23/06/2003
    !
    !-------------------------------------------
    implicit none
    real ( kind = pr ), intent(in) :: f(:,:), h
    real ( kind = pr ) :: res(size(f,1))
    integer :: nn, i
    !
    nn = size(f,2)
    if ( nn == 2 * ( nn / 2 ) ) stop 'Simpson...'
    res(:) = f(:,1) - f(:,nn)
    do i = 2, nn - 1, 2
       res(:) = res(:) + 4 * f(:,i) + 2 * f(:,i+1)
    end do
    res = h / 3 * res
  end function vecsimpson

  !
  !!
  !

  function rombergs( f, h ) result(res)
    !-------------------------------------------
    !
    ! Purpose:
    !  Numerical using Romberg method for a scalar real function.
    ! Input:
    !  f = 1 dimension array of real( kind = pr )
    !  h = step
    ! Output:
    !  res = integral of f
    !
    ! Certified: 23/06/2003
    !
    !-------------------------------------------
    implicit none
    integer :: pwr
    real ( kind = pr ) :: f(0:), res
    real ( kind = pr ) :: ab, h
    real ( kind = pr ), dimension(:,:), allocatable :: t
    integer mmax, nmax
    integer :: coe
    integer :: m, i, ni, j, ifail, n2m1

    ni = size(f) - 1
    pwr = int(1.e-7_pr+log( real( ni, kind = pr ) ) / log(2.0_pr) )
    if ( ni /= 2**pwr ) then
       print *, '** Incorrect vector dimension for Romberg (s)...'
       print *, ni, 2**pwr
       stop
    end if
    allocate( t( pwr+1, pwr+1 ), stat = ifail )
    if ( ifail /= 0 ) stop '** Not enough memory...'
    ab = h * ni
    nmax = pwr + 1
    mmax = pwr + 1
    t(1,1) = ( f(0) + f(ni) ) * ab / 2
    do m = 2, mmax
       t(m,1) = 0.0_pr
       n2m1 = ni / 2**(m-1)
       do i = 1, 2**(m-1) - 1, 2
          j = i * n2m1
          t(m,1) = t(m,1) + f(j)
       end do
       t(m,1) = t(m-1,1) / 2 + ab * t(m,1) / 2**(m-1)
    end do
    do ni = 2, nmax
       coe = 4**(ni-1)
       do m = 1, mmax + 1 - ni
          t(m,ni) = ( coe * t(m+1,ni-1) - t(m,ni-1) ) / ( coe - 1 )
       end do
    end do
    res = t(1,nmax)
    deallocate(t)
    return
  end function rombergs

  !
  !!
  !

  function rombergv( f, h ) result(res)
    !-------------------------------------------
    !
    ! Purpose:
    !  Numerical using Romberg method for a vector real function.
    ! Input:
    !  f = n dimension array of real( kind = pr )
    !  h = step
    ! Output:
    !  res = n dimensions array, integral of f
    !
    ! Certified: 23/06/2003
    !
    !-------------------------------------------
    implicit none
    integer :: pwr
    real ( kind = pr ) :: f(:,0:), res(1:size(f,1))
    real ( kind = pr ) :: ab, h
    real ( kind = pr ), dimension(:,:,:), allocatable :: t
    integer mmax, nmax
    integer :: coe
    integer :: m, i, ni, j, ifail, n2m1

    ni = size(f,2) - 1
    pwr = int(1.e-7_pr+log( real( ni, kind = pr ) ) / log(2.0_pr) )
    if ( ni /= 2**pwr ) then
       print *, '** Incorrect vector dimension for Romberg (v)...'
       print *, ni, 2**pwr
       stop
    end if
    allocate( t( size(f,1), pwr+1, pwr+1 ), stat = ifail )
    if ( ifail /= 0 ) stop '** Not enough memory...'
    ab = ni * h
    nmax = pwr + 1
    mmax = pwr + 1
    t(:,1,1) = ( f(:,0) + f(:,ni) ) * ab / 2
    do m = 2, mmax
       t(:,m,1) = 0.0_pr
       n2m1 = ni / 2**(m-1)
       do i = 1, 2**(m-1) - 1, 2
          j = i * n2m1
          t(:,m,1) = t(:,m,1) + f(:,j)
       end do
       t(:,m,1) = t(:,m-1,1) / 2 + ab * t(:,m,1) / 2**(m-1)
    end do
    do ni = 2, nmax
       coe = 4**(ni-1)
       do m = 1, mmax + 1 - ni
          t(:,m,ni) = ( coe * t(:,m+1,ni-1) - t(:,m,ni-1) ) / ( coe - 1 )
       end do
    end do
    res(:) = t(:,1,nmax)
    deallocate(t)
    return
  end function rombergv

  !
  !!
  !

end module sinm_difint
module sinm_forces
  use sinm_cste
  use sinm_skyrme
  implicit none

  !
  !!
  !

  real ( kind = pr ), private :: c0a, c0b, c12a, c12b, cm, cmi
  real ( kind = pr ) :: wa0, wa1, wb0, wb1, ja, jb
  real ( kind = pr ) :: j_0, j_1 ! same as ja, jb but in T form
  real ( kind = pr ) :: c3aa, c3ba, c3ca, &
       &                c3ab, c3bb, c3cb, &
       &                c3ac, c3bc, c3cc
  real ( kind = pr ), private :: af1, af2, af5, af6, af7, af8
  real ( kind = pr ), private :: af3a, af4a, af3b, af4b, af3c, af4c

  !
  !!
  !

  interface computefields
     module procedure computefields
  end interface

  interface diffuseness
     module procedure diffuseness
  end interface

  interface compute_energy
     module procedure compute_energy
  end interface

  interface compute_sigs
     module procedure compute_sigs
  end interface

  interface setforce
     module procedure setforce
  end interface

  interface print_out_energies
     module procedure print_out_energies
  end interface

contains

  subroutine setforce()
    !-------------------------------------------
    !
    ! Purpose:
    ! Input:
    ! Output:
    !
    ! Certified:
    !
    !-------------------------------------------
    implicit none
    !
    !
    !! Usefull coef for:
    !
    !! ... effective mass
    !
    cm = ( t1 * ( 2 + x1 ) + t2 * ( 2 + x2 ) ) / 8
    cmi = ( t2 * ( 2 * x2 + 1 ) - t1 * ( 2 * x1 + 1 ) ) / 8
    !
    !! ... mean field
    !
    c0a = t0 * ( 2 + x0 ) / 2
    c0b = - t0 * ( 2 * x0 + 1 ) / 2
    !
    c3aa = t3a * ( 2 + x3a ) * ( 2 + gama ) / 24 
    c3ba = - t3a * ( 2 * x3a + 1 ) / 12
    c3ca = - t3a * gama * ( 2 * x3a + 1 ) / 24
    !
    c3ab = t3b * ( 2 + x3b ) * ( 2 + gamb ) / 24 
    c3bb = - t3b * ( 2 * x3b + 1 ) / 12
    c3cb = - t3b * gamb * ( 2 * x3b + 1 ) / 24
    !
    c3ac = t3c * ( 2 + x3c ) * ( 2 + gamc ) / 24 
    c3bc = - t3c * ( 2 * x3c + 1 ) / 12
    c3cc = - t3c * gamc * ( 2 * x3c + 1 ) / 24
    !
    c12a = ( t2 * ( 2 + x2 ) - 3 * t1 * ( 2 + x1 ) ) / 16
    c12b = ( 3 * t1 * ( 2 * x1 + 1 ) + t2 * ( 2 * x2 + 1 ) ) / 16
    !
    !! ... Spin orbit
    !
    wa0 = - w0 / 2
    wa1 = - w1 / 2
    wb0 = w0 / 2
    wb1 = w1 / 2
    !
    !! ... terms coming from the J**2 contributions
    !
    ja = 0.0_pr
    jb = 0.0_pr
    j_0 = 0.0_pr
    j_1 = 0.0_pr
    if ( j2terms ) then
       ja = - ( t1 * x1 + t2 * x2 ) / 8 + 5.0_pr / 24 * ( tt + tu )
       jb = ( t1 - t2 ) / 8             + 5.0_pr / 24 * ( tu - tt )
       j_0 = t1 / 32 * ( 1 - 2 * x1 ) - t2 / 32 * ( 1 + 2 * x2 )  &
            + 5 / 32.0_pr * ( tt / 3 + tu )
       j_1 = ( t1 - t2 ) / 32                                     &
            - 5 / 32.0_pr * ( tt / 3 - tu / 3 )
    end if
    !
    af1 = t0 * ( 2 + x0 ) / 4.0_pr
    af2 = - t0 * ( 2 * x0 + 1 ) / 4.0_pr
    af3a = t3a * ( 2 + x3a ) / 24.0_pr
    af4a = - t3a * ( 2 * x3a + 1 ) / 24.0_pr
    af3b = t3b * ( 2 + x3b ) / 24.0_pr
    af4b = - t3b * ( 2 * x3b + 1 ) / 24.0_pr
    af3c = t3c * ( 2 + x3c ) / 24.0_pr
    af4c = - t3c * ( 2 * x3c + 1 ) / 24.0_pr
    af5 = ( t1 * ( 2 + x1 ) + t2 * ( 2 + x2 ) ) / 8.0_pr
    af6 = - ( 3 * t1 * ( 2 + x1 ) - t2 * ( 2 + x2 ) ) / 32.0_pr
    af7 = - ( t1 * ( 2 * x1 + 1 ) - t2 * ( 2 * x2 + 1 ) ) / 8.0_pr
    af8 = ( 3 * t1 * ( 2 * x1 + 1 ) + t2 * ( 2 * x2 + 1 ) ) / 32.0_pr
    !stop "reverifier les termes pour le tenseur !!!"
    !
  end subroutine setforce

  !
  !!
  !

  subroutine computefields(first)
    !-------------------------------------------
    !
    ! Purpose:
    ! Input:
    ! Output:
    !
    ! Certified:
    !
    !-------------------------------------------
    use sinm_system
    use sinm_geometry
    implicit none
    logical, intent(in) :: first
    integer :: iso
    real ( kind = pr ) :: x
    !
    x = 1.0_pr - xf
    if ( first ) then
       hb2m   = 0.0_pr
       dhb2m  = 0.0_pr
       d2hb2m = 0.0_pr
       u_cent = 0.0_pr
       u_so   = 0.0_pr
       x = 1.0_pr
    end if
    !
    do iso = 1, 2
       !
       !! Effective mass
       !
       if ( yp == 0.5_pr ) then

          hb2m(iso,:) = ( 1.0_pr - x ) * hb2m(iso,:)              &
               + x * ( hb2 + thetas / 16 * rho(3,:) )
          dhb2m(iso,:) = ( 1.0_pr - x ) * dhb2m(iso,:)            &
               + x * ( thetas / 16 * drho(3,:) )
          d2hb2m(iso,:) = ( 1.0_pr - x ) * d2hb2m(iso,:)          &
               + x * ( thetas / 16 * d2rho(3,:) )
       else
          hb2m(iso,:) = ( 1.0_pr - x ) * hb2m(iso,:)              &
               + x * ( hb2 + cm * rho(3,:) + cmi * rho(iso,:) )
          dhb2m(iso,:) = ( 1.0_pr - x ) * dhb2m(iso,:)            &
               + x * ( cm * drho(3,:)  + cmi * drho(iso,:)  )
          d2hb2m(iso,:) = ( 1.0_pr - x ) * d2hb2m(iso,:)          &
               + x * ( cm * d2rho(3,:) + cmi * d2rho(iso,:) )
       end if
       !
       !! Central field
       !
       if ( yp == 0.5_pr ) then
          u_cent(iso,:) = ( 1.0_pr - x ) * u_cent(iso,:)                 &
               &        + x * (                                          &
               &     c0a * rho(3,-npt:npt)   +  c0b * rho(iso,-npt:npt)  &
               &  +   cm * tau(3,-npt:npt)   +  cmi * tau(iso,-npt:npt)  &
               &  + ( thetas - 12 * t1 ) / 32 * d2rho(3,-npt:npt)     &
               &  +  c3aa * rho(3,-npt:npt)**(gama+1)                    &! a
               &  +  c3ba * rho(3,-npt:npt)**gama * rho(iso,-npt:npt)    &! a
               &  +  c3ca * rho(3,-npt:npt)**(gama-1)                    &! a
               &         * ( rho(1,-npt:npt)**2 + rho(2,-npt:npt)**2 )   &! a
               &  +  c3ab * rho(3,-npt:npt)**(gamb+1)                    &! b
               &  +  c3bb * rho(3,-npt:npt)**gamb * rho(iso,-npt:npt)    &! b
               &  +  c3cb * rho(3,-npt:npt)**(gamb-1)                    &! b
               &         * ( rho(1,-npt:npt)**2 + rho(2,-npt:npt)**2 )   &! b
               &  +  c3ac * rho(3,-npt:npt)**(gamc+1)                    &! c
               &  +  c3bc * rho(3,-npt:npt)**gamc * rho(iso,-npt:npt)    &! c
               &  +  c3cc * rho(3,-npt:npt)**(gamc-1)                    &! c
               &         * ( rho(1,-npt:npt)**2 + rho(2,-npt:npt)**2 )   &! c
               &  +   wa0 * dspin(3,-npt:npt) + wa1 * dspin(iso,-npt:npt) )
       else
          u_cent(iso,:) = ( 1.0_pr - x ) * u_cent(iso,:)                  &
               &        + x * (                                           &
               &     c0a * rho(3,-npt:npt)   +  c0b * rho(iso,-npt:npt)   &
               &  +   cm * tau(3,-npt:npt)   +  cmi * tau(iso,-npt:npt)   &
               &  + c12a * d2rho(3,-npt:npt) + c12b * d2rho(iso,-npt:npt) &
               &  +  c3aa * rho(3,-npt:npt)**(gama+1)                    &! a
               &  +  c3ba * rho(3,-npt:npt)**gama * rho(iso,-npt:npt)    &! a
               &  +  c3ca * rho(3,-npt:npt)**(gama-1)                    &! a
               &         * ( rho(1,-npt:npt)**2 + rho(2,-npt:npt)**2 )   &! a
               &  +  c3ab * rho(3,-npt:npt)**(gamb+1)                   &! b
               &  +  c3bb * rho(3,-npt:npt)**gamb * rho(iso,-npt:npt)   &! b
               &  +  c3cb * rho(3,-npt:npt)**(gamb-1)                   &! b
               &         * ( rho(1,-npt:npt)**2 + rho(2,-npt:npt)**2 )  &! b
               &  +  c3ac * rho(3,-npt:npt)**(gamc+1)                    &! c
               &  +  c3bc * rho(3,-npt:npt)**gamc * rho(iso,-npt:npt)    &! c
               &  +  c3cc * rho(3,-npt:npt)**(gamc-1)                    &! c
               &         * ( rho(1,-npt:npt)**2 + rho(2,-npt:npt)**2 )   &! c
               &  +   wa0 * dspin(3,-npt:npt) + wa1 * dspin(iso,-npt:npt) )
       end if
       !
       !! Spin-Orbit field
       !
       u_so(iso,:) = ( 1.0_pr - x ) * u_so(iso,:)                        &
            & + x * (                                                    &
            &          wb0 * drho(3,-npt:npt) + wb1 * drho(iso,-npt:npt) &
            &        + ja * spin(3,-npt:npt) + jb * spin(iso,-npt:npt) )
    end do
    !
  end subroutine computefields

  !
  !!
  !

  subroutine compute_energy( total_energy, ef, ekin, eso )
    !-------------------------------------------
    !
    ! Purpose:
    ! Input:
    ! Output:
    !
    ! Certified:
    !
    !-------------------------------------------
    use sinm_difint
    use sinm_geometry
    use sinm_system
    implicit none
    real ( kind = pr ) :: total_energy, ef, ekin(3), eso
    integer :: i
    !
    !! Kinetic energy
    !
    do i = 1, 2
       energy(i,-npt:npt) = hb2 * tau(i,-npt:npt)
    end do
    !
    !! Field energy
    !
    energy(3,-npt:npt) =                                                &
         &   af1 * rho(3,-npt:npt)**2                                   &
         & + af2 * ( rho(1,-npt:npt)**2 + rho(2,-npt:npt)**2 )
    !
    energy(5,-npt:npt) = rho(3,-npt:npt)**2   &
         &       * ( af3a * rho(3,-npt:npt)**gama )                  &! a
         &         + ( rho(1,-npt:npt)**2 + rho(2,-npt:npt)**2 )     &! a
         &                        * ( af4a * rho(3,-npt:npt)**gama )
    !
    energy(6,-npt:npt) = rho(3,-npt:npt)**2   &
         &       * ( af3b * rho(3,-npt:npt)**gamb )                  &! b
         &         + ( rho(1,-npt:npt)**2 + rho(2,-npt:npt)**2 )     &! b
         &                        * ( af4b * rho(3,-npt:npt)**gamb )
    !
    energy(7,-npt:npt) = rho(3,-npt:npt)**2   &
         &       * ( af3c * rho(3,-npt:npt)**gamc )                  &! c
         &         + ( rho(1,-npt:npt)**2 + rho(2,-npt:npt)**2 )     &! c
         &                        * ( af4c * rho(3,-npt:npt)**gamc )
    !
    if ( yp == 0.5_pr ) then
       energy(4,-npt:npt) =                                       &
            &    tau(3,-npt:npt) * rho(3,-npt:npt)* thetas / 16   &
            &  - drho(3,-npt:npt)**2 / 64 * ( thetas - 12 * t1 )
    else
       energy(4,-npt:npt) = af5 * rho(3,-npt:npt) * tau(3,-npt:npt)     &
            & - af6 * drho(3,-npt:npt)**2                               &
            & + af7 * ( rho(1,-npt:npt) * tau(1,-npt:npt)               &
            &                    + rho(2,-npt:npt) * tau(2,-npt:npt) )  &
            & - af8 * ( drho(1,-npt:npt)**2 + drho(2,-npt:npt)**2 )
    end if
    if ( j2terms ) then
       !energy(4,-npt:npt) = energy(4,-npt:npt)                         &
       !     & - ( t1 * x1 + t2 * x2 ) / 16 * spin(3,-npt:npt)**2       &
       !     & + ( t1 - t2 ) / 16                                       &
       !     &       * ( spin(1,-npt:npt)**2 + spin(2,-npt:npt)**2 )
       energy(4,-npt:npt) = energy(4,-npt:npt)                         &
            & + j_0 * spin(3,-npt:npt)**2       &
            & + j_1 * ( spin(1,-npt:npt) - spin(2,-npt:npt) )**2
    end if
    !
    !! Spin-orbit energy
    !
    !energy(8,-npt:npt) = - w0 / 2 * ( rho(3,-npt:npt) * dspin(3,-npt:npt)   &
    !     &                      + rho(1,-npt:npt) * dspin(1,-npt:npt)       &
    !     &                      + rho(2,-npt:npt) * dspin(1,-npt:npt) )

    energy(8,-npt:npt) = wa0 * rho(3,-npt:npt) * dspin(3,-npt:npt)         &
         &             + wa1 * ( rho(1,-npt:npt) * dspin(1,-npt:npt)       &
         &                     + rho(2,-npt:npt) * dspin(2,-npt:npt) )
    !
    !! Total
    !
    energy(9,:) = energy(1,:) + energy(2,:) + energy(3,:)  &
         &      + energy(4,:)                              &
         &      + energy(5,:) + energy(6,:) + energy(7,:)  &
         &      + energy(8,:)
    !
    !if ( integration == "r" ) then
    !   ekin(1) = romberg( energy(1,-npt:npt), h )
    !   ekin(2) = romberg( energy(2,-npt:npt), h )
    !   ef      = romberg( energy(3,-npt:npt), h )  &
    !        &  + romberg( energy(4,-npt:npt), h )  &
    !        &  + romberg( energy(5,-npt:npt), h )  &
    !        &  + romberg( energy(6,-npt:npt), h )  &
    !        &  + romberg( energy(7,-npt:npt), h )
    !   eso     = romberg( energy(8,-npt:npt), h )
    !else
    ekin(1) = simpson( energy(1,-npt:npt) )
    ekin(2) = simpson( energy(2,-npt:npt) )
    !ef      = simpson( energy(3,-npt:npt) )  &
    !     &  + simpson( energy(4,-npt:npt) )  &
    !     &  + simpson( energy(5,-npt:npt) )  &
    !     &  + simpson( energy(6,-npt:npt) )  &
    !     &  + simpson( energy(7,-npt:npt) )
    ef      = simpson( energy(3,-npt:npt)    &
         &  +          energy(4,-npt:npt)    &
         &  +          energy(5,-npt:npt)    &
         &  +          energy(6,-npt:npt)    &
         &  +          energy(7,-npt:npt) )
    eso     = simpson( energy(8,-npt:npt) )
    !end if
    !
    ekin(3) = ekin(1) + ekin(2)
    total_energy = ekin(3) + ef + eso
    !
  end subroutine compute_energy

  !
  !!
  !

  subroutine print_out_energies(prefix)
    use sinm_difint
    use sinm_geometry
    use sinm_system
    implicit none
    character ( len = 3 ), intent(in) :: prefix
    integer :: i
    real ( kind = pr ) :: rho_0, rho_1, tau_0, tau_1
    real ( kind = pr ) :: rd2r_0, rd2r_1, dj_0, dj_1, j2_0, j2_1
    real ( kind = pr ) :: c_rho_0, c_rho_1, c_tau_0, c_tau_1
    real ( kind = pr ) :: c_drho_0, c_drho_1, c_dj_0, c_dj_1
    real ( kind = pr ) :: c_j2_0, c_j2_1
    real ( kind = pr ), parameter :: hf = 0.5_pr, cq = 1.25_pr
    !
    return
    open( unit = 55, file = "out/"//prefix//"_energies.gfx", &
         status = "unknown" )
    open( unit = 56, file = "out/"//prefix//"_energiesdd.gfx", &
         status = "unknown" )
    do i = - npt, npt
       rho_0 = rho(1,i) + rho(2,i)
       rho_1 = - ( rho(1,i) - rho(2,i) )
       tau_0 = tau(1,i) + tau(2,i)
       tau_1 = - ( tau(1,i) - tau(2,i) )
       rd2r_0 = ( rho(1,i) + rho(2,i) ) * ( d2rho(1,i) + d2rho(2,i) )
       rd2r_1 = - ( rho(1,i) - rho(2,i) ) * ( d2rho(1,i) - d2rho(2,i) )
       dj_0 = dspin(1,i) + dspin(2,i)
       dj_1 = - ( dspin(1,i) - dspin(2,i) ) !  FAUX !!!!!!!!
       j2_0 = ( spin(1,i) + spin(2,i) )**2
       j2_1 = ( spin(1,i) - spin(2,i) )**2
       !
       c_rho_0 = 3.0_pr / 8 * t0               &
            + 3.0_pr / 48 * t3a * rho_0**gama  &
            + 3.0_pr / 48 * t3b * rho_0**gamb  &
            + 3.0_pr / 48 * t3c * rho_0**gamc
       c_rho_1 = - t0 / 4 * ( hf + x0 )              &
            - t3a / 24 * ( hf + x3a ) * rho_0**gama  &
            - t3b / 24 * ( hf + x3b ) * rho_0**gamb  &
            - t3c / 24 * ( hf + x3c ) * rho_0**gamc
       !
       c_tau_0 = 3.0_pr / 16 * t1 + t2 / 4 * ( cq + x2 )
       c_tau_1 = - t1 / 8 * ( hf + x1 ) + t2 / 8 * ( hf + x2 )
       !
       c_drho_0 = - 9.0_pr / 64 * t1 + t2 / 16 * ( cq + x2 )
       c_drho_1 = 3.0_pr / 32 * t1 * ( hf + x1 ) + t2 / 32 * ( hf + x2 )
       !
       if ( j2terms ) then
          !c_j2_0 = t1 / 8 * ( hf - x1 ) - t2 / 8 * ( hf + x2 )
          c_j2_0 =  2 * j_0
          !c_j2_1 = ( t1 - t2 ) / 16
          c_j2_1 = 2 * j_1
       else
          c_j2_0 = 0.0_pr
          c_j2_1 = 0.0_pr
       end if
       !
       if ( .not. xtls ) then
          c_dj_0 = - 3.0_pr / 4 * w0
          c_dj_1 = - 1.0_pr / 4 * w0
       else
          c_dj_0 = - w0 / 2 - w1 / 4
          c_dj_1 = - w1 / 4
       end if
       !
       write( 55, '(f7.3,12e13.5)' ) z(i),                           &
            energy(1,i) + energy(2,i),                               &
            - ( energy(1,i) - energy(2,i) ),                         &
            c_rho_0 * rho_0**2,      c_rho_1 * rho_1**2,             &
            c_tau_0 * rho_0 * tau_0, c_tau_1 * rho_1 * tau_1,        &
            c_drho_0 * rd2r_0,       c_drho_1 * rd2r_1,              &
            c_j2_0 * j2_0,           c_j2_1 * j2_1,                  &
            c_dj_0 * rho_0 * dj_0,   c_dj_1 * rho_1 * dj_1
       write( 56, '(f7.3,12e13.5)' ) z(i),                      &
            3.0_pr / 8 * t0 * rho_0**2,                         &
            - t0 / 4 * ( hf + x0 ) * rho_1**2,                  &
            3.0_pr / 48 * t3a * rho_0**gama * rho_0**2,         &
            - t3a / 24 * ( hf + x3a ) * rho_0**gama * rho_1**2, &
            3.0_pr / 48 * t3b * rho_0**gamb * rho_0**2,         &
            - t3b / 24 * ( hf + x3b ) * rho_0**gamb * rho_1**2, &
            3.0_pr / 48 * t3c * rho_0**gamc * rho_0**2,         &
            - t3c / 24 * ( hf + x3c ) * rho_0**gamc * rho_1**2
    end do
    close(55)
    close(56)
    !
  end subroutine print_out_energies

  !
  !!
  !

  subroutine compute_sigs( sig_t, sig_0, sig_12, &
       sig_3a, sig_3b, sig_3c, sig_w, vir, d2sig )
    use sinm_difint
    use sinm_geometry
    use sinm_system
    implicit none
    real ( kind = pr ), intent(out) :: sig_t, sig_0, sig_12, &
         sig_3a, sig_3b, sig_3c, sig_w, d2sig, vir
    !if ( integration == "r" ) then
    !   sig_t  = romberg( energy(1,-npt:npt) + energy(2,-npt:npt), h )
    !   sig_0  = romberg( energy(3,-npt:npt), h )
    !   sig_12 = romberg( energy(4,-npt:npt), h )
    !   sig_3a = romberg( energy(5,-npt:npt), h )
    !   sig_3b = romberg( energy(6,-npt:npt), h )
    !   sig_3c = romberg( energy(7,-npt:npt), h )
    !   sig_w  = romberg( energy(8,-npt:npt), h )
    !else
    sig_t  = simpson( energy(1,-npt:npt) + energy(2,-npt:npt) )  &
         - 0.6_pr * hb2 * dtpi2**t23 * simpson( rho(3,-npt:npt) ) &
         * rho0(3)**(t23)
    sig_0  = simpson( energy(3,-npt:npt) )  &
         - 3.0_pr / 8 * t0 * rho0(3) * simpson( rho(3,-npt:npt) )
    sig_12 = simpson( energy(4,-npt:npt) )  &
         - 3.0_pr / 80 * thetas * dtpi2**t23 &
         * rho0(3)**t53 * simpson( rho(3,-npt:npt) )
    sig_3a = simpson( energy(5,-npt:npt) )  &
         - t3a / 16 * rho0(3)**(gama+1) * simpson( rho(3,-npt:npt) )
    sig_3b = simpson( energy(6,-npt:npt) )  &
         - t3b / 16 * simpson( rho(3,-npt:npt)**(gamb+1) )
    sig_3c = simpson( energy(7,-npt:npt) )  &
         - t3c / 16 * simpson( rho(3,-npt:npt)**(gamc+1) )
    sig_w  = simpson( energy(8,-npt:npt) )
    !end if
    !
    !    print *, simpson( energy(1,-n:n) + energy(2,-n:n), h )
    !    print '("sig_t  = ",f12.6)', sig_t
    !    print '("sig_0  = ",f12.6)', sig_0
    !    print '("sig_12 = ",f12.6)', sig_12
    !    print '("sig_w  = ",f12.6)', sig_w
    !    print '("sig_3a = ",f12.6)', sig_3a
    !    print '("sig_3b = ",f12.6)', sig_3b
    !    print '("sig_3c = ",f12.6)', sig_3c
    !    print '("sig    = ",f12.6)', sig_t + sig_0 + sig_12 + sig_w  &
    !         + sig_3a + sig_3b + sig_3c
    !print '("s(",i4,")** = ",f12.3)', n, ( 20 * sig_t + 30 * sig_0 &
    !+ 56 * ( sig_12 + sig_w ) &
    !+ ( 3 * gama + 5 ) * ( 3 * gama + 6 ) * sig_3a ) / 9 / rho0(3)**2
    vir = rho0(3) * ( 4 * sig_t + 5 * sig_0 + 7 * sig_12 + 7 * sig_w  &
         + ( 3 * gama + 5 ) * sig_3a )                                &
         / ( sig_t + sig_0 + sig_12 + sig_w + sig_3a )
    d2sig = ( 20 * sig_t + 30 * sig_0 &
         + 56 * ( sig_12 + sig_w ) &
         + ( 3 * gama + 5 ) * ( 3 * gama + 6 ) * sig_3a ) / 9 / rho0(3)**2
    !    !sig_t  = 4 * sig_t
    !    !sig_0  = 5 * sig_0
    !    !sig_12 = 7 * sig_12
    !    !sig_w  = 7 * sig_w
    !    !sig_3a = ( 3 * gama + 5 ) * sig_3a
    !    !sig_3b = ( 3 * gamb + 5 ) * sig_3b
    !    !sig_3c = ( 3 * gamc + 5 ) * sig_3c
    !    !
  end subroutine compute_sigs

  !
  !!
  !

  function diffuseness() result(res)
    !-------------------------------------------
    !
    ! Purpose:
    !  Compute the surface diffuseness of the total density with
    !  the definition of W. Myers in Nucl. Phys. A 204 (1973) 465.
    ! Input:
    ! Output:
    !
    ! Certified: 24/06/2003
    !
    !-------------------------------------------
    use sinm_difint
    use sinm_geometry
    use sinm_system
    implicit none
    real ( kind = pr ) :: res
    real ( kind = pr ) :: g(-npt:npt), c, b2
    integer :: n0
    !
    g(-npt:npt) = drho(3,-npt:npt)
    n0 = - npt
    do
       if ( g(n0) * g(n0+1) < 0.0_pr ) exit
       n0 = n0 + 1
    end do
    g(n0+1:npt) = 0.0_pr
    !if ( integration == "r" ) then
    !   g = g / romberg( g, h )
    !   c = romberg( g(-npt:npt) * z(-npt:npt), h )
    !   b2 = romberg( g(-npt:npt) * ( z(-npt:npt) - c )**2, h )
    !else
    g = g / simpson(g)
    c = simpson( g(-npt:npt) * z(-npt:npt) )
    b2 = simpson( g(-npt:npt) * ( z(-npt:npt) - c )**2 )
    !end if
    res = sqrt(b2)
    !
  end function diffuseness

  !
  !!
  !

  function diffuseness_t() result(res)
    use sinm_system
    use sinm_geometry
    implicit none
    real ( kind = pr ) :: res(3)
    real ( kind = pr ) :: z0(3), z1(3)
    integer :: i, k
    logical :: ifail
    !
    do k = 1, 3
       do i = 0, npt
          ifail = .false.
          if ( rho(k,i) > 0.9_pr * rho0(k) ) exit
          ifail = .true.
       end do
       z1(k) = z(i-1) + ( z(i) - z(i-1) ) / ( rho(k,i) - rho(k,i-1) ) &
            * ( 0.9_pr * rho0(k) - rho(k,i-1) )
       do i = 0, - npt, -1
          ifail = .false.
          if ( rho(k,i) < 0.1_pr * rho0(k) ) exit
          ifail = .true.
       end do
       z0(k) = z(i) + ( z(i+1) - z(i) ) / ( rho(k,i+1) - rho(k,i) ) &
            * ( 0.1_pr * rho0(k) - rho(k,i) )
    end do
    res = z1 - z0
    !
  end function diffuseness_t

  !
  !!
  !

end module sinm_forces
module sinm_fields
  use sinm_cste
  use sinm_skyrme
  use sinm_system
  use sinm_geometry
  implicit none

  interface init
     module procedure init
  end interface

contains

  subroutine init()
    implicit none
    real ( kind = pr ), parameter :: d = 0.54_pr
    real ( kind = pr ) :: f53
    integer :: i
    !
    do i = - n3, n3
       rho(1:2,i) = rho0(1:2) / ( 1 + exp( - z(i) / d ) )
    end do
    rho(3,:) = rho(1,:) + rho(2,:)
    !
    !!
    !
    kf(1) = ( 3 * pi**2 * rho0(1) )**t13
    kf(2) = ( 3 * pi**2 * rho0(2) )**t13
    kf(3) = ( 1.5_pr * pi**2 * rho0(3) )**t13
    !
    f53 = 2.0_pr**t23 * ( yp**t53 + ( 1 - yp )**t53 )
    tau(1,:) = ( 3 * pi**2 * rho(1,:) )**t53 / 5 / pi**2
    tau(2,:) = ( 3 * pi**2 * rho(2,:) )**t53 / 5 / pi**2
    tau(3,:) = tau(1,:) + tau(2,:)
    !
    spin = 0.0_pr
    !
  end subroutine init

end module sinm_fields
module sinm_eqdif
  use sinm_cste
  implicit none

  interface inflexion_point
     module procedure inflexion_point
  end interface

  interface shift_densities
     module procedure shift_densities
  end interface

contains

  !
  !!
  !

  subroutine inflexion_point( ninf, epsilon )
    !-------------------------------------------
    !
    ! Purpose:
    !  Determine the position of the total density inflexion point.
    ! Input:
    ! Output:
    !  ninf = integer
    !  epsilon = real( kind = pr )
    !  inflexion point is at  ninf * h + epsilon
    !
    ! Certified: 23/06/2003
    !
    !-------------------------------------------
    use sinm_geometry
    implicit none
    real ( kind = pr ) :: epsilon
    integer :: ninf
    integer :: ni(1), i
    real ( kind = pr ) :: d, da, db, dc, z1, z2, z3, f1, f2, f3, &
         a, b, c, delta
    !
    do ninf = - npt + 10, npt - 10
       if ( d2rho(3,ninf) * d2rho(3,ninf+1) <= 0.0_pr ) exit
    end do
    ni = minloc(abs(d2rho(3,ninf-1:ninf+1)))
    ninf = ninf + ni(1) - 2
    if ( abs(d2rho(3,ninf)) >= abs(d2rho(3,ninf+1)) ) ninf = ninf + 1
    if ( ninf >= npt - 10 ) stop "No inflexion point in density..."
    z1 = z(ninf-1)
    z2 = z(ninf)
    z3 = z(ninf+1)
    f1 = d2rho(3,ninf-1)
    f2 = d2rho(3,ninf)
    f3 = d2rho(3,ninf+1)
    d = z2 * z3 * ( z3 - z2 ) + z1 * z3 * ( z1 - z3 ) &
         + z1 * z2 * ( z2 - z1 )
    da = f1 * z2 * z3 * ( z3 - z2 ) + f2 * z1 * z3 * ( z1 - z3 ) &
         + f3 * z1 * z2 * ( z2 - z1 )
    db = ( f2 * z3**2 - f3 * z2**2 ) - ( f1 * z3**2 - f3 * z1**2 ) &
         + ( f1 * z2**2 - f2 * z1**2 )
    dc = ( z2 * f3 - z3 * f2 ) - ( z1 * f3 - z3 * f1 ) &
         + ( z1 * f2 - z2 * f1 )
    epsilon =  &
         - 2 * h * d2rho(3,ninf) / ( d2rho(3,ninf+1) - d2rho(3,ninf-1) )
    a = da / d
    b = db / d
    c = dc / d
    delta = b * b - 4 * a * c
    epsilon = ( - b + sqrt(delta) ) / 2 / c
    if ( abs( ( - b - sqrt(delta) ) / 2 / c ) < abs(epsilon) ) &
         epsilon = ( - b - sqrt(delta) ) / 2 / c
    if ( abs(epsilon) >= 6*h ) then
       print *, 'epsilon =', epsilon
       print *, 'ninf = ', ninf
       print *, d2rho(3,ninf-1:ninf+1)
       open( unit = 55, file = "ooops", status = "unknown" )
       do i = - npt, npt
          write( 55, '(f7.3,3e18.6)' ) z(i), rho(:,i)
       end do
       close(55)
       epsilon = 0.0_pr
       ninf = 0
       !stop
    end if
    epsilon = epsilon - ninf * h
    !
  end subroutine inflexion_point

  !
  !!
  !

  subroutine shift_densities( ns, epsilon )
    !-------------------------------------------
    !
    ! Purpose:
    !  Shift densities
    ! Input:
    !  ns = integer: algebraic number of points for the shift
    !  epsilon = real( kind = pr ) remaining shift
    ! Output:
    !
    ! Certified: 23/06/2003
    !
    !-------------------------------------------
    use sinm_difint
    use sinm_geometry
    implicit none
    integer, intent(in) :: ns
    real ( kind = pr ), intent(in) :: epsilon
    integer :: i, ne
    !
    if ( ns /= 0 ) then
       if ( ns > 0 ) then
          ne = npt
       else
          ne = - npt
       end if
       do i = 1, 3
          rho(i,:)    = eoshift( rho(i,:),    ns, rho(i,ne)    )
          drho(i,:)   = eoshift( drho(i,:),   ns, drho(i,ne)   )
          d2rho(i,:)  = eoshift( d2rho(i,:),  ns, d2rho(i,ne)  )
          tau(i,:)    = eoshift( tau(i,:),    ns, tau(i,ne)    )
          dtau(i,:)   = eoshift( dtau(i,:),   ns, dtau(i,ne)   )
          d2tau(i,:)  = eoshift( d2tau(i,:),  ns, d2tau(i,ne)  )
          spin(i,:)   = eoshift( spin(i,:),   ns, spin(i,ne)   )
          dspin(i,:)  = eoshift( dspin(i,:),  ns, dspin(i,ne)  )
          d2spin(i,:) = eoshift( d2spin(i,:), ns, d2spin(i,ne) )
       end do
    end if
    if ( abs(epsilon) /= 0.0_pr ) then
       do i = - npt, npt
          rho(:,i) = rho(:,i) + epsilon * drho(:,i) &
               + epsilon**2 / 2.0_pr * d2rho(:,i)
          tau(:,i) = tau(:,i) + epsilon * dtau(:,i) &
               + epsilon**2 / 2.0_pr * d2tau(:,i)
          spin(:,i) = spin(:,i) + epsilon * dspin(:,i) &
               + epsilon**2 / 2.0_pr * d2spin(:,i)
       end do
       !
       ne = npt - ns
       call deriv( rho(:,-ne:ne),  drho(:,-ne:ne),  d2rho(:,-ne:ne)  )
       call deriv( tau(:,-ne:ne),  dtau(:,-ne:ne),  d2tau(:,-ne:ne)  )
       call deriv( spin(:,-ne:ne), dspin(:,-ne:ne), d2spin(:,-ne:ne) )
    end if

  end subroutine shift_densities

  !
  !!
  !

end module sinm_eqdif
module sinm_hf
  use sinm_cste
  use sinm_geometry
  use sinm_fields
  !use io
  implicit none

  real ( kind = pr ), dimension(:), allocatable :: ueff, uc
  real ( kind = pr ), dimension(:,:), allocatable :: &
       rhot, taut, spint, rhoz, tauz, spinz
  real ( kind = pr ), dimension(:), allocatable :: fm, fps, fpc, &
       f, df, ff, d2ff

  interface sky
     module procedure sky
  end interface

contains

  subroutine inisky()
    implicit none
    integer :: ifail
    allocate( uc(-npt:npt), ueff(-npt:npt),                             &
         fm(-n3:npt), fps(-npt:n3), fpc(-npt:n3),                       &
         f(-n3:n3), df(-n3:n3), ff(-npt:npt), d2ff(-npt:npt),           &
         rhot(-n3:n3,0:nkt),  taut(-n3:n3,0:nkt), spint(-n3:n3,0:nkt),  &
         rhoz(-n3:n3,0:nkz),  tauz(-n3:n3,0:nkz), spinz(-n3:n3,0:nkz),  &
         stat = ifail )
    if ( ifail /= 0 ) stop
  end subroutine inisky

  !
  !!
  !

  subroutine stopsky()
    implicit none
    deallocate( uc, ueff, fm, fps, fpc, f, rhot, df, taut, &
         rhoz, tauz, spinz, spint, ff, d2ff )
  end subroutine stopsky

  !
  !!
  !

  subroutine sky( iso, lambda )
    use sinm_difint
    implicit none
    integer, intent(in) :: iso, lambda
    integer :: iz, it
    real ( kind = pr ) :: hz, ht, lq, l, t, kt, kt2, kz, en, g, a
    real ( kind = pr ) :: dfs, dfc, dfm, x, y, det, dx, dy, xe
    logical :: prob
    !
    uc =    0.5_pr * d2hb2m(iso,-npt:npt)                          &
         - 0.25_pr * dhb2m(iso,-npt:npt)**2 / hb2m(iso,-npt:npt)   &
         + u_cent(iso,:)
    lq = kf(iso) / kf(3)
    hz = lq / nkz
    xe = 1.0_pr
    !
    do
       prob = .false.
       rhoz  = 0.0_pr
       tauz  = 0.0_pr
       spinz = 0.0_pr
       !
       do iz = 1, nkz - 1
          !
          l = iz * hz
          kz = l * kf(3) * xe
          if ( lq**2 - l**2 < 0.0_pr ) stop 'hmmm...'
          ht = sqrt( lq**2 - l**2 ) / nkt
          rhot  = 0.0_pr
          taut  = 0.0_pr
          spint = 0.0_pr
          !
          do it = 1, nkt
             t = ht * it
             kt = t * kf(3) * xe
             kt2 = kt * kt
             en = hb2m(iso,npt) * ( kz**2 + kt2 ) + uc(npt) &
                  + lambda * kt * u_so(iso,npt)
             ! ??
             if ( kt2 - en / hb2m(iso,npt) <= 0.0_pr ) then
                print *, kt2 - en / hb2m(iso,-npt)
                print *, - kz**2 - &
                     ( uc(-npt) + lambda * kt * u_so(iso,-npt) ) &
                     / hb2m(iso,-npt)
                print *, - kz**2 - uc(npt) / hb2m(iso,npt)
                print *, 'argg...'
                prob = .true.
                stop
                exit
             end if
             !
             g = sqrt( kt2 - en / hb2m(iso,npt) )
             ueff = hh12 * ( ( uc + lambda * kt * u_so(iso,:) - en ) &
                  &          / hb2m(iso,-npt:npt) + kt2 )
             !
             !! Forward integration
             !
             fm(-n3:-npt+1) = exp( g * z(-n3:-npt+1) )
             call numerov( fm(-npt:3), ueff(-npt:3), 1 )
             !
             !! Backward intergration
             !
             fps(npt-1:n3) = sin( kz * z(npt-1:n3) )  &
                  * sqrt( hb2m(iso,npt-1:n3) )
             call numerov( fps(-3:npt), ueff(-3:npt), -1 )
             !
             fpc(npt-1:n3) = cos( kz * z(npt-1:n3) )  &
                  * sqrt( hb2m(iso,npt-1:n3) )
             call numerov( fpc(-3:npt), ueff(-3:npt), -1 )
             !
             dfs = ( 45 * ( fps(1) - fps(-1) ) - 9 * ( fps(2) - fps(-2) )  &
                  &       + fps(3) - fps(-3) ) / h_60
             dfc = ( 45 * ( fpc(1) - fpc(-1) ) - 9 * ( fpc(2) - fpc(-2) )  &
                  &       + fpc(3) - fpc(-3) ) / h_60
             dfm = ( 45 * ( fm(1) - fm(-1) ) - 9 * ( fm(2) - fm(-2) )      &
                  &       + fm(3) - fm(-3) ) / h_60
             det = fps(0) * dfc - fpc(0) * dfs
             !
             if ( abs(det) < 1.e-50_pr ) then
                print *, 'Determinant is zero...'
                print '(" kz =",f7.3,3x,"kt =",f7.3,3x,"g =",f7.3)', kz, kt, g
                stop
             else
                dx = fm(0) * dfc - fpc(0) * dfm
                dy = fps(0) * dfm - fm(0) * dfs
                x = dx / det
                y = dy / det
                a = sqrt( ( x * x + y * y ) )
                x = x / a
                y = y / a
             end if
             !
             f(0:n3) = fps(0:n3) * x + fpc(0:n3) * y
             f(-n3:-1) = fm(-n3:-1) / a
             !
             f  = f / sqrt( hb2m(iso,:) )
             !
             call deriv( f, df )
             !
             rhot(:,it)  = abs(f)**2 * t
             taut(:,it)  = ( kf(3)**2 * t**2 * abs(f)**2 + abs(df)**2 ) * t
             spint(:,it) = abs(f)**2 * t**2
             !
          end do
          !
          if ( prob ) exit
          !
          !if ( integration == "r" ) then
          !   rhoz(:,iz)  = romberg( rhot,  ht )
          !   tauz(:,iz)  = romberg( taut,  ht )
          !   spinz(:,iz) = romberg( spint, ht )
          !else
          rhoz(:,iz)  = simpson( rhot,  ht )
          tauz(:,iz)  = simpson( taut,  ht )
          spinz(:,iz) = simpson( spint, ht )
          !end if
          !
       end do
       !
       if ( prob ) then
          xe = 0.95_pr
          print *, "warning..."
          cycle
       else
          exit
       end if
    end do
    !
    !if ( integration == "r" ) then
    !   rho(iso,:)  = rho(iso,:)  + romberg( rhoz,  hz )
    !   tau(iso,:)  = tau(iso,:)  + romberg( tauz,  hz )
    !   spin(iso,:) = spin(iso,:) + romberg( spinz, hz ) * lambda
    !else
    rho(iso,:)  = rho(iso,:)  + simpson( rhoz,  hz )
    tau(iso,:)  = tau(iso,:)  + simpson( tauz,  hz )
    spin(iso,:) = spin(iso,:) + simpson( spinz, hz ) * lambda
    !end if
    !
  end subroutine sky

  !
  !!
  !

  subroutine numerov( f, u, s )
    implicit none
    real ( kind = pr ), intent(in) :: u(:)
    real ( kind = pr ), intent(inout) :: f(:)
    integer, intent(in) :: s
    integer :: n, i, n1, n2
    !
    n = size(u)
    n1 = 2
    n2 = n - 1
    if ( s < 0 ) then
       n1 = n - 1
       n2 = 2
    end if
    do i = n1, n2, s
       f(i+s) = ( ( 2 + 10 * u(i) ) * f(i)     &
            &     - ( 1 - u(i-s) ) * f(i-s) )  &
            &  / ( 1 - u(i+s) )
    end do
    !
  end subroutine numerov

  !
  !!
  !

end module sinm_hf
module sinm_sinm
  use sinm_cste
  implicit none
  real ( kind = pr ) :: as_av

  interface get_sinm
     module procedure get_sinm
  end interface

contains

  subroutine get_sinm( av, as, sig, vir, d2sig, b, t, nucnum, etot )
    use sinm_forces
    use sinm_fields
    use sinm_io
    use sinm_eqdif
    use sinm_hf
    use sinm_difint
    implicit none
    real ( kind = pr ), intent(in) :: av
    real ( kind = pr ), intent(out) :: as, sig, d2sig, vir, b, nucnum(3), &
         etot, t(3)
    !
    integer :: iter, iso, lambda, i, ninf, n_av
    real ( kind = pr ) :: e0, e1, ep, sigma, asf, epsilon, ekin(3), &
         ef, de, eso, sig_t, sig_0, sig_12, sig_w,       &
         sig_3a, sig_3b, sig_3c
    logical, parameter :: debug = .false.
    logical :: last
    !
    call init()
    call deriv( rho, drho,  d2rho  )
    call deriv( tau, dtau,  d2tau  )
    !
    !! Start iterations
    !
    call inisky()
    e0 = 0.0_pr
    ninf = 0
    !
    print '(" iter    Kinetic Energy         Field  ",  &
         &"   Eso      Total         D(E)")'
    print '("        neutr.  prot.   Tot.")'
    !
    last = .false.
    !
    do iter = 1, 999
       !
       call computefields( iter == 1 )
       !
       !! Flush densisties
       !
       rho  = 0.0_pr
       tau  = 0.0_pr
       spin = 0.0_pr
       energy = 0.0_pr
       !
       do iso = 1, 2
          !
          if ( iso == 2 .and. yp == 0.5_pr ) then
             rho(2,:)  = rho(1,:)
             tau(2,:)  = tau(1,:)
             spin(2,:) = spin(1,:)
          else
             !
             do lambda = -1, 1, 2
                call sky( iso, lambda )
             end do
             !
             rho(iso,:)  = 1.5_pr * rho0(3) * rho(iso,:)
             tau(iso,:)  = 1.5_pr * rho0(3) * tau(iso,:)
             spin(iso,:) = 1.5_pr * rho0(3) * spin(iso,:) * kf(3)
             !
          end if
          !
       end do ! iso
       !
       rho(3,:)  = rho(1,:)  + rho(2,:)
       tau(3,:)  = tau(1,:)  + tau(2,:)
       spin(3,:) = spin(1,:) + spin(2,:)
       call deriv( rho, drho,  d2rho  )
       call deriv( tau, dtau,  d2tau  )
       call deriv( spin, dspin, d2spin )
       !
       call inflexion_point( ninf, epsilon )
       if ( ninf > 3  ) ninf =  3  ! don't shift too
       if ( ninf < -3 ) ninf = -3  ! much at once...
       call shift_densities( ninf, epsilon )
       !
       call compute_energy( e1, ef, ekin, eso )
       !
       !if ( integration == "r" ) then
       !   e1 = romberg( energy(9,-npt:npt), h )
       !else
       e1 = simpson( energy(9,-npt:npt) )
       !end if
       !
       de = abs( ( e1 - e0 ) / e1 )
       print '(1x,i3,1x,3f8.3,2(1x,f8.3),1x,f10.5,1x,e12.4)', &
            iter, ekin, ef, eso, e1, de
       e0 = e1
       if ( last ) exit
       if ( de < 1.e-8_pr ) last = .true.
       !
    end do
    !
    !! End of iterations
    !
    !if ( integration == "r" ) then
    !   ep = av * romberg( rho(3,-npt:npt), h )
    !else
    ep = av * simpson( rho(3,-npt:npt) )
    !end if
    sigma = e1 - ep
    asf = ( 36 * pi / rho0(3)**2 )**(1.0_pr/3.0_pr) * sigma
    call compute_sigs( sig_t, sig_0, sig_12, &
         sig_3a, sig_3b, sig_3c, sig_w, vir, d2sig )
    !sig = sig_t + sig_0 + sig_12 + sig_w + sig_3a + sig_3b + sig_3c
    sig = sigma ! the upper formula is wrong in asymetric matter, why ?
    !
    !!
    !
    as = asf
    b = diffuseness()
    t =  diffuseness_t()
    if ( integration == "r" ) then
       nucnum = romberg( rho(:,-npt:npt), h )
    else
       nucnum = simpson( rho(:,-npt:npt), h )
    end if
    etot = e1 / nucnum(3)
    !
    !call line()
    !
    !! Print out densities and fields
    !
    !if ( gfxo ) then
    !open( unit = 20, file = 'out/nrj.gfx', status = 'unknown' )
    !do i = -npt, npt
    !   write(20,'(f7.3,3e16.6e3,f9.6)') i * h, av * rho(3,i), &
    !        &  energy(9,i), &
    !        &  energy(9,i) - av * rho(3,i)
    !end do
    !close(20)
    n_av = 0
    as_av = 0.0_pr
    open( unit = 20, file = 'out/chk.gfx', status = 'unknown' )
    write( 20, '("#  z",10x,"E",12x,"a_v * rho",6x,"E - av * rho",6x,&
         &"a_s(z)")' ) 
    do i = -npt, npt
       asf = ( sum( energy(9,-npt:i) - av * rho(3,-npt:i) ) &
            - ( energy(9,i) - av * rho(3,i) ) / 2 ) * h
       asf = ( 36 * pi / rho0(3)**2 )**(1.0_pr/3.0_pr) * asf
       write(20,'(f7.3,3e16.6e3,f12.6)') i * h, energy(9,i), &
            &  av * rho(3,i), &
            &  energy(9,i) - av * rho(3,i), asf
       if ( i * h >= 4 * t(3) ) then
          as_av = as_av + asf
          n_av = n_av + 1
       end if
    end do
    close(20)
    if ( n_av == 0 ) stop "The box seems to be too small..."
    as_av = as_av / n_av
    !   end if
    !   if ( gfxo ) then
    call printoutdensities()
    !      call printoutfields()
    !   end if
    !
    !! End
    !
    call stopsky()
    !   call freemem()

  end subroutine get_sinm

end module sinm_sinm
program main
  use sinm_cste
  use sinm_skyrme
  use sinm_inm
  use sinm_sinm
  use sinm_system
  use sinm_io
  use sinm_geometry
  use sinm_forces
  implicit none
  real ( kind = pr ) :: av, kinf, d3e, ai, r0, as, sig, d2sig, vir
  real ( kind = pr ) :: h0, h1, v0, v1, b, nucnum(3), etot, ksf, t(3)
  real ( kind = pr ) :: rbox_start, rbox_stop, yp_start, yp_stop, rbox
  real ( kind = pr ) :: delta_z
  !real ( kind = pr ), dimension(:), allocatable :: a_surf
  integer :: i, irun, irbox, iyp, irm
  character ( len = name_length ) :: skforce, force
  character ( len = 2 ) :: skr
  character ( len = 3 ) :: prefix
  character ( len = 30 ) :: dummy
  character ( len = 1 ) :: head
  integer :: iprefix, ifail
  logical :: lvir, lso, ltens
  namelist /input/ rbox_start, rbox_stop, irbox, yp_start, yp_stop,  &
       iyp, lso, force, lvir, m_kz, ltens
  !
  !!
  !
  delta_z = 0.15_pr
  call setcste()
  open( unit = 12, file = 'out/results.dat', status = 'unknown' )
  write( 12, &
       '("# npt   R_box    Y_p      I      a_surf  <a_surf>  K_inf    t_p",&
       &"    t_n    t")' )
  !
  lvir = .false.
  lso = .true.
  ltens = .true.
  force = ""
  rbox_start = 12.0_pr
  rbox_stop  = 12.0_pr
  irbox = 1
  yp_start = 0.5_pr
  yp_stop  = 94.0_pr / 240.0_pr
  iyp = 5
  m_kz = 2
  !
  open( unit = 14, file = "sinm.input", status = "old" )
  read( 14, nml = input )
  close(14)
  if ( iyp /= 1 .and. irbox /= 1 )  &
       stop "Ne pas faire varier yp et Rbox en meme temps..."
  call read_force( trim(adjustl(force)) )
  skforce = trim(adjustl(force))
  !
  h = delta_z
  irm = max( iyp, irbox ) - 1
  rbox = rbox_start
  yp = yp_start
  !
  do irun = 0, irm
     !
     iprefix = irun
     if ( irun == 0 ) then
        prefix = "000"
     else
        if ( irun < 10 ) then
           write( prefix, '("00",i1)' ) iprefix
        else
           if ( irun < 100 ) then
              write( prefix, '("0",i2)' ) iprefix
           else
              if ( irun < 1000 ) then
                 write( prefix, '(i3)' ) iprefix
              else
                 stop "too many files..."
              end if
           end if
        end if
     end if
     !
     if ( irbox > 1 ) then
        rbox = rbox_start + ( irun * ( rbox_stop - rbox_start ) ) / irm
     else
        rbox = rbox_start
     end if
     if ( iyp > 1 ) then
        yp = yp_start + ( irun * ( yp_stop - yp_start ) ) / irm
     end if
     npt = int( rbox / delta_z ) !+ 1
     rbox = npt * h
     !
     if ( .not.xtls ) w1 = w0
     if ( .not. lso ) then
        w0 = 0.0_pr
        w1 = 0.0_pr
     end if
     if ( .not. ltens ) then
        tt = 0.0_pr
        tu = 0.0_pr
     end if
     !
     !!
     !

     call set_theta()
     call line()
     print '(20x,5("-"),3x,a,3x,5("-"))', skforce
     call print_param()
     call line()

     !
     !!
     !
     call get_inm( yp, av, ai, kinf, d3e, r0 )
     call print_inm( av, ai, kinf, r0 )

     call line()

     !
     !!
     !
     open( unit = 30, file = 'out/vir.gfx', status = 'unknown' )
     i = 0
     do
        !npt = 150

        !npt = irun

        if ( i == 0 ) then
           h = delta_z
        else
           if ( i == 1 ) then
              h = delta_z * 1.1_pr !0.098_pr
           else
              h = ( v1 * h0 - v0 * h1 ) / ( v1 - v0 )
              if ( h < 0.0_pr .or. h > 0.29_pr ) h = ( h0 + h1 ) / 2
           end if
        end if
        nkt = 4 ! 8  * 2
        nkz = m_kz ! 32 * m_kz ! 32 ?
        xf = 0.75_pr
        call set_geometry()
        call print_geo()
        call line()

        gfx = .false.
        call setforce()
        call alloc_mem()
        !
        call get_sinm( av, as, sig, vir, d2sig, b, t, nucnum, etot )
        !
        ksf = 4 * pi * r0**2 * ( 22 * sig + 9 * rho0(3)**2 * d2sig  &
             + 54 * sig / kinf * rho0(3)**3 * d3e )
        call line()
        print '(" | rho0  = ",f7.4," + ",f7.4," = ",f6.3," fm-3")', rho0
        print '(" | Energy per nucleon =",f8.3," MeV")', etot
        print '(" | Surface energy =",f8.3," MeV,",6x,"===>  Averaged = ",&
             &f8.3," MeV")', as, as_av
print '(" |",42x,23("-"))'
        print '(" | Diffuseness    = ",f5.2," fm  (b parameter of Myers)")', b
        print '(" | Thickness (90%-10%):   t_p =",f5.2," fm,   t_n =",&
             &f5.2," fm,   t =",f5.2," fm")', t
        print '(" | Nucleons number =",f7.3," +",f7.3," =",f7.3," fm-2")', &
             nucnum
        if ( yp == 0.5_pr ) then
           print '(" | Sigma =",f8.4," MeV.fm-2",5x,"d2sig = ",f8.2,&
                &" Mev.fm4")', sig, d2sig
           print '(" | h = ",f9.6,4x,"vir = ",f11.8,4x,"R_box = ",f6.3)', &
                h, vir, npt * h
           print '(" | K_inf = ",f7.2,4x,"K_sf = ",f8.2)', kinf, ksf
        else
           print '(" | K_inf = ",f7.2)', kinf
           print '(" | Sigma =",f8.4," MeV.fm-2")', sig
        end if
        call line()
        !
        if ( yp /= 0.5_pr .or. .not. lvir ) then
           call print_out_energies(prefix)
           call free_mem()
           exit
        end if
        if ( i == 0 ) then
           v0 = vir
           h0 = h
        else
           if ( i == 1 ) then
              v1 = vir
              h1 = h
           else
              h0 = h1
              h1 = h
              v0 = v1
              v1 = vir
           end if
        end if
        write( 30, '(4e18.8)' ) vir, sig, d2sig, ksf
        call free_mem()
        i = i + 1

        !i = 999

        if ( i > 20 .or. abs(vir) < 0.000001_pr ) exit
     end do
     !
     write( 12, '(1x,i4,f8.3,f8.4,f8.5,1x,3f9.4,3f7.3)' ) &
          npt, npt * h, yp, &
          1 - 2 * yp, as, as_av, kinf, t
     print '(18x,a,3x," .... finished",///)', skforce
     close(30)
     !
  end do
  close(12)
  !
  !head = " "
  !allocate( a_surf(0:irm), stat = ifail )
  !if ( ifail /= 0 ) stop "Can not allocate memory for a_surf..."
  !open( unit = 12, file = "out/results.dat", status = "old" )
  !open( unit = 14, file = "out/results.prob", status = "unknown" )
  !read( 12, '(a)' ) prefix
  !do i = 0, irm
  !   if ( i == irm ) head = "#"
  !   read( 12, '(a30,f9.4)' ) dummy, a_surf(i)
  !   write( 14, '(a1,a30,i5,2f9.4)' ) head, dummy, nkz, a_surf(i),  &
  !        sum( a_surf(0:i) ) / ( i + 1 )
  !end do
  !close(14)
  !close(12)
  !deallocate( a_surf )
  !
end program main
